__major_version__  = 0
__minor_version__  = 6
__patch__          = 8
__version_tuple__ = (
    __major_version__,
    __minor_version__,
    __patch__
)
__version_number__ = ".".join([str(v) for v in __version_tuple__])
__version__ = f"{__version_number__}"
